"""Other smaller Enums for OpenDrift"""

from enum import Enum


# Enum for drift_model
class DriftModelEnum(str, Enum):
    """Enum for drift models used in OpenDrift."""

    OceanDrift = "OceanDrift"
    LarvalFish = "LarvalFish"
    OpenOil = "OpenOil"
    Leeway = "Leeway"
    HarmfulAlgalBloom = "HarmfulAlgalBloom"


# Enum for radius_type
class RadiusTypeEnum(str, Enum):
    """Enum for radius types used in OpenDrift."""

    gaussian = "gaussian"
    uniform = "uniform"


# Define Pydantic Enum classes
class DiffusivityModelEnum(str, Enum):
    """Enum for diffusivity models used in OpenDrift."""

    environment = "environment"
    stepfunction = "stepfunction"
    windspeed_Sundby1983 = "windspeed_Sundby1983"
    windspeed_Large1994 = "windspeed_Large1994"
    gls_tke = "gls_tke"
    constant = "constant"


class CoastlineActionEnum(str, Enum):
    """Enum for coastline actions used in OpenDrift."""

    none = "none"
    stranding = "stranding"
    previous = "previous"


class SeafloorActionEnum(str, Enum):
    """Enum for seafloor actions used in OpenDrift."""

    none = "none"
    lift_to_seafloor = "lift_to_seafloor"
    deactivate = "deactivate"
    previous = "previous"


class PlotTypeEnum(str, Enum):
    """Enum for plot types used in OpenDrift."""

    spaghetti = "spaghetti"
    animation = "animation"
    animation_profile = "animation_profile"
    oil = "oil"
    property = "property"
    all = "all"


class DropletSizeDistributionEnum(str, Enum):
    """Enum for droplet size distribution types."""

    uniform = "uniform"
    normal = "normal"
    lognormal = "lognormal"



"""Test realistic scenarios, which are slower."""

import pickle

from pathlib import Path

import numpy as np
import pytest
import xarray as xr

import particle_tracking_manager as ptm


# set up an alternate dataset on-the-fly
ds = xr.Dataset(
    data_vars={
        "u": (("ocean_time", "Z", "Y", "X"), np.zeros((2, 3, 2, 3))),
        "v": (("ocean_time", "Z", "Y", "X"), np.zeros((2, 3, 2, 3))),
        "w": (("ocean_time", "Z", "Y", "X"), np.zeros((2, 3, 2, 3))),
        "salt": (("ocean_time", "Z", "Y", "X"), np.zeros((2, 3, 2, 3))),
        "temp": (("ocean_time", "Z", "Y", "X"), np.zeros((2, 3, 2, 3))),
        "wetdry_mask_rho": (("ocean_time", "Z", "Y", "X"), np.zeros((2, 3, 2, 3))),
        "mask_rho": (("Y", "X"), np.zeros((2, 3))),
        "Uwind": (("ocean_time", "Y", "X"), np.zeros((2, 2, 3))),
        "Vwind": (("ocean_time", "Y", "X"), np.zeros((2, 2, 3))),
        "Cs_r": (("Z"), np.linspace(-1, 0, 3)),
        "hc": 16,
    },
    coords={
        "ocean_time": ("ocean_time", [0, 1], {"units": "seconds since 1970-01-01"}),
        "s_rho": (("Z"), np.linspace(-1, 0, 3)),
        "lon_rho": (("Y", "X"), np.array([[1, 2, 3], [1, 2, 3]])),
        "lat_rho": (("Y", "X"), np.array([[1, 1, 1], [2, 2, 2]])),
    },
)
ds_info = dict(
    lon_min=1, lon_max=3, lat_min=1, lat_max=2, start_time_model=0, end_time_fixed=1
)

ptm.config_ocean_model.register_on_the_fly(ds_info)


# also to use the user-defined template of the TXLA model, need to input where pooch is downloading
# the file
ptm.config_ocean_model.update_TXLA_with_download_location()


@pytest.mark.slow
def test_add_new_reader():
    """Add a separate reader from the defaults using ds."""

    manager = ptm.OpenDriftModel(
        steps=1, ocean_model="ONTHEFLY", lon=2, lat=1.5, start_time=0, time_step=0.01
    )
    manager.add_reader(ds=ds)


@pytest.mark.slow
def test_run_parquet():
    """Set up and run."""

    seeding_kwargs = dict(lon=-90, lat=28.7, number=1, start_time="2009-11-19T12:00:00")
    manager = ptm.OpenDriftModel(
        **seeding_kwargs,
        use_static_masks=True,
        steps=2,
        output_format="parquet",
        ocean_model="TXLA",
        ocean_model_local=False,
    )
    manager.run_all()

    assert "parquet" in manager.o.outfile_name


@pytest.mark.slow
def test_run_netcdf_and_plot():
    """Set up and run."""

    import tempfile

    ts = 6 * 60  # 6 minutes in seconds

    seeding_kwargs = dict(lon=-90, lat=28.7, number=1, start_time="2009-11-19T12:00:00")
    with tempfile.NamedTemporaryFile(delete=False) as temp_file:
        manager = ptm.OpenDriftModel(
            **seeding_kwargs,
            use_static_masks=True,
            steps=2,
            output_format="netcdf",
            use_cache=True,
            interpolator_filename=temp_file.name,
            ocean_model="TXLA",
            ocean_model_local=False,
            plots={
                "all": {},
            },
            time_step=ts,
        )
        manager.run_all()

        assert "nc" in manager.o.outfile_name
        assert manager.config.interpolator_filename == Path(temp_file.name).with_suffix(
            ".pickle"
        )

        # Replace 'path_to_pickle_file.pkl' with the actual path to your pickle file
        with open(manager.config.interpolator_filename, "rb") as file:
            data = pickle.load(file)
        assert "spl_x" in data
        assert "spl_y" in data

    # check time_step across access points
    assert (
        # m.o._config["general:time_step_minutes"]["value"]  # this is not correct, don't know why
        manager.o.time_step.total_seconds()
        == ts
        == manager.config.time_step
        # == m.o.get_configspec()["general:time_step_minutes"]["value"]  # this is not correct, don't know why
    )


@pytest.mark.slow
def test_run_HarmfulAlgalBloom_biomass_change():
    """Set up and run HarmfulAlgalBloom and match biomass change."""

    seeding_kwargs = dict(lon=-90, lat=28.7, number=1, start_time="2009-11-19T12:00:00")
    manager = ptm.OpenDriftModel(
        **seeding_kwargs,
        use_static_masks=True,
        duration="1h",
        ocean_model="TXLA",
        ocean_model_local=False,
        drift_model="HarmfulAlgalBloom",
    )
    manager.run_all()

    # check that biomass decreased due to temperature-induced mortality
    # calculated as: biomass = initial_biomass * exp(-mortality_rate_high * time)
    # initial biomass is 1.0, mortality_rate_high is 0.5 days^-1, time is 1 hour = 1/24 days
    assert np.allclose(
        float(manager.o.elements.biomass[0]),
        np.exp(-manager.config.mortality_rate_high * 3600 / 86400),
    )



"""Defines OceanModelConfig with classes stored in ocean_model_registry.

Set up ocean model configuration: doesn't depend on a tracking simulation.
OceanModelConfig instances contain information about ocean models that is relevant to the model itself, separate from a particle tracking simulation.
"""

import itertools

# Standard library imports
import os
import pprint

from datetime import datetime
from pathlib import Path

# Third-party imports
import pandas as pd
import xarray as xr
import yaml

from pydantic import BaseModel, Field
from typing_extensions import Annotated


def calculate_CIOFSOP_max() -> datetime:
    """read in CIOFSOP max time available, as datetime object"""
    try:
        date = (
            xr.open_dataset(
                "/mnt/depot/data/packrat/prod/noaa/coops/ofs/aws_ciofs/processed/aws_ciofs_kerchunk.parq",
                engine="kerchunk",
            )
            .ocean_time[-1]
            .values.astype("datetime64[s]")
            .item()
        )
    except:
        date = (pd.Timestamp.now() + pd.Timedelta("1d")).isoformat()
    return date


def get_model_end_time(name: str) -> datetime:
    """Get the end time of the model based on its name."""
    # This is only run when the property is requested
    if name == "CIOFSOP":
        return calculate_CIOFSOP_max()
    else:
        raise NotImplementedError(f"get_model_end_time not implemented for {name}.")


class OceanModelConfig(BaseModel):
    """Ocean model configuration."""

    name: Annotated[
        str,
        Field(description="Name of the model."),
    ]
    temporal_resolution_str: Annotated[
        str,
        Field(
            description="ISO 8601 format temporal resolution of the model. e.g. 'PT1H' for hourly resolution."
        ),
    ]
    lon_min: Annotated[
        float,
        Field(description="Minimum longitude of the model."),
    ]
    lon_max: Annotated[
        float,
        Field(description="Maximum longitude of the model."),
    ]
    lat_min: Annotated[
        float,
        Field(description="Minimum latitude of the model."),
    ]
    lat_max: Annotated[
        float,
        Field(description="Maximum latitude of the model."),
    ]
    start_time_model: Annotated[
        datetime,
        Field(description="Start time of the model."),
    ]
    oceanmodel_lon0_360: Annotated[
        bool,
        Field(
            description="Set to True to use 0-360 longitude convention for this model."
        ),
    ]
    standard_name_mapping: Annotated[
        dict[str, str],
        Field(description="Mapping of model variable names to standard names."),
    ]
    model_drop_vars: Annotated[
        list[str],
        Field(
            description="List of variables to drop from the model dataset. These variables are not needed for particle tracking."
        ),
    ]
    loc_remote: Annotated[
        str | None,
        Field(description="Remote location of the model dataset."),
    ]
    chunks: Annotated[
        dict | None,
        Field(description="Chunking strategy for the model dataset."),
    ]
    dx: Annotated[
        float | None,
        Field(
            description="Approximate horizontal grid resolution (meters), used to calculate horizontal diffusivity."
        ),
    ]

    end_time_fixed: Annotated[
        datetime | None,
        Field(None, description="End time of the model, if doesn't change."),
    ]

    kerchunk_func_str: Annotated[
        str | None,
        Field(
            description="Name of function to create a kerchunk file for the model, mapped to function name in function_map."
        ),
    ]

    @property
    def end_time_model(self) -> datetime:
        """Get the end time of the model."""
        if self.end_time_fixed:
            return self.end_time_fixed
        else:  # there is only one that uses this currently
            return get_model_end_time(self.name)

    @property
    def horizontal_diffusivity(self) -> float | None:
        """Calculate horizontal diffusivity based on known ocean_model.

        Might be overwritten by user-input in other model config.
        """

        if self.dx is None:
            return None

        # horizontal diffusivity is calculated based on the mean horizontal grid resolution
        # for the model being used.
        # 0.1 is a guess for the magnitude of velocity being missed in the models, the sub-gridscale velocity
        sub_gridscale_velocity = 0.1
        horizontal_diffusivity = sub_gridscale_velocity * self.dx
        return horizontal_diffusivity


class OceanModelRegistry:
    """Registry for OceanModelConfig instances."""

    def __init__(self):
        """Initialize the OceanModelRegistry."""
        self._registry = {}

    def __repr__(self):
        """Return a string representation of the registry."""
        return str(self.all())

    def register(self, name: str, instance: OceanModelConfig):
        """Register an OceanModelConfig instance by its name."""
        self._registry[name] = instance

    def get(self, name: str):
        """Get an OceanModelConfig instance by its name."""
        return self._registry.get(name)

    def show(self, name: str):
        """Show the details of a registered OceanModelConfig instance."""
        return pprint.pprint(self._registry[name].model_dump())

    def get_all(self):
        """Return all registered OceanModelConfig instances."""
        return self._registry.items()

    def all(self):
        """Return all registered OceanModelConfig instance names."""
        return list(self._registry.keys())

    def all_models(self):
        """Return all registered OceanModelConfig instances."""
        return list(self._registry.values())

    def update_model(self, name, changes: dict):
        """Update a registered OceanModelConfig instance with new values."""
        if name in self._registry:
            for key, value in changes.items():
                setattr(self._registry[name], key, value)
        else:
            raise ValueError(f"Model {name} not found in registry.")


# Directory with YAML files
directory = (
    Path(__file__).resolve().parent / "ocean_models"
)  # This is the directory where the current script is located

file_paths: list = list(directory.glob("*.yaml"))

# Directory with user-defined files, if any
config_dir = Path(os.getenv("PTM_CONFIG_DIR", ""))
if len(str(config_dir)) > 1:
    file_paths = list(itertools.chain(file_paths, config_dir.glob("*.yaml")))

# also combine *.yaml files in the user_ocean_models directory specifically (not just by default)
config_dir = Path(__file__).resolve().parent / "user_ocean_models"
file_paths = list(itertools.chain(file_paths, config_dir.glob("*.yaml")))

# Create an instance of the OceanModelRegistry
ocean_model_registry = OceanModelRegistry()

# Iterate through all .yaml files in the directory
for file_path in file_paths:
    with open(file_path, "r") as f:
        config_data = yaml.safe_load(f)[file_path.stem]

        # Assuming your config_data needs to be loaded into a Pydantic model
        # Create the OceanModelConfig instance from the data
        config = OceanModelConfig(**config_data)

        # Register the configuration, perhaps by its name
        ocean_model_registry.register(config.name, config)



from datetime import datetime

import pandas as pd
import pytest

from pydantic import ValidationError

from particle_tracking_manager.config_the_manager import TheManagerConfig


def test_start_time_type():
    """Check start time type."""
    m = TheManagerConfig(steps=1, start_time="2022-01-01 12:00:00")
    assert m.start_time == pd.Timestamp("2022-01-01 12:00:00")

    m = TheManagerConfig(steps=1, start_time=pd.Timestamp("2022-01-01 12:00:00"))
    assert m.start_time == pd.Timestamp("2022-01-01 12:00:00")

    m = TheManagerConfig(steps=1, start_time=datetime(2022, 1, 1, 12, 0, 0))
    assert m.start_time == pd.Timestamp("2022-01-01 12:00:00")


def test_time_calculations():
    with pytest.raises(ValidationError):
        m = TheManagerConfig(steps=1, start_time=None)

    with pytest.raises(ValidationError):
        m = TheManagerConfig(steps=1, duration="1d", start_time=None)

    with pytest.raises(ValidationError):
        # all times defined but not consistently
        m = TheManagerConfig(
            steps=1,
            duration="1d",
            start_time="2022-01-01 12:00:00",
            end_time="2022-01-03 12:00:00",
        )

    # all times defined but consistently
    # duration is 1 time step
    ts, duration = 5 * 60, "P0DT0H5M0S"
    start_time = "2022-01-01 12:00:00"
    m = TheManagerConfig(
        time_step=ts,
        steps=1,
        duration=duration,
        start_time=start_time,
        end_time=pd.Timestamp(start_time) + pd.Timedelta(seconds=ts),
    )

    m = TheManagerConfig(steps=1, end_time="2022-01-01 12:00:00", start_time=None)
    assert m.duration == duration
    assert m.start_time == m.end_time - pd.Timedelta(m.duration)

    with pytest.raises(ValidationError):
        m = TheManagerConfig(
            steps=1,
            end_time="2000-01-02",
            start_time=pd.Timestamp("2000-1-1"),
            ocean_model="CIOFS",
        )

    m = TheManagerConfig(
        end_time="2000-01-02", start_time=pd.Timestamp("2000-1-1"), ocean_model="CIOFS"
    )
    assert m.steps == 288
    assert m.duration == pd.Timedelta("1 days 00:00:00").isoformat()

    m = TheManagerConfig(
        end_time="2023-01-02", start_time=pd.Timestamp("2023-1-1"), run_forward=True
    )
    assert m.timedir == 1

    m = TheManagerConfig(
        end_time="2023-01-02", start_time=pd.Timestamp("2023-1-1"), run_forward=False
    )
    assert m.timedir == -1


def test_lon_lat():
    """Check for valid lon and lat values

    In the general sense from TheManagerConfig.
    Checked for individual ocean_models in test_config_ocean_model.py.
    """

    with pytest.raises(ValidationError):
        m = TheManagerConfig(steps=1, start_time="2022-01-01", lon=-180.1)

    with pytest.raises(ValidationError):
        m = TheManagerConfig(steps=1, start_time="2022-01-01", lat=95)

    m = TheManagerConfig(steps=1, lon=-152, lat=58, start_time="2022-01-01")
    assert m.lon == -152
    assert m.lat == 58


def test_unknown_parameter():
    """Make sure unknown parameters are not input."""

    with pytest.raises(ValidationError):
        m = TheManagerConfig(unknown="test", steps=1, start_time="2022-01-01")


def test_horizontal_diffusivity_logic():
    """Check logic for using default horizontal diff values for known models."""

    m = TheManagerConfig(ocean_model="NWGOA", steps=1, start_time="2007-01-01")
    assert m.horizontal_diffusivity == 150.0  # known grid values

    m = TheManagerConfig(ocean_model="CIOFS", steps=1, start_time="2020-01-01")
    assert m.horizontal_diffusivity == 10.0  # known grid values

    m = TheManagerConfig(ocean_model="CIOFSOP", horizontal_diffusivity=11, steps=1)
    assert m.horizontal_diffusivity == 11.0  # user-selected value

    m = TheManagerConfig(ocean_model="CIOFSOP", steps=1)
    assert m.horizontal_diffusivity == 10.0  # known grid values


def test_z():
    m = TheManagerConfig(steps=1, start_time="2022-01-01", z=-10)

    with pytest.raises(ValueError):
        m = TheManagerConfig(steps=1, start_time="2022-01-01", z=10)


def test_seed_flag_elements():
    """Check seed flag elements."""

    with pytest.raises(ValidationError):
        m = TheManagerConfig(
            steps=1, start_time="2022-01-01", seed_flag="elements", lon=None, lat=None
        )

    m = TheManagerConfig(steps=1, start_time="2022-01-01", seed_flag="elements")


def test_seed_flag_geojson():
    """Check seed flag geojson."""
    geojson = {
        "type": "Feature",
        "properties": {},
        "geometry": {"type": "Point", "coordinates": [0, 0]},
    }

    with pytest.raises(ValidationError):
        m = TheManagerConfig(
            steps=1, start_time="2022-01-01", seed_flag="geojson", geojson=None
        )

    with pytest.raises(ValidationError):
        m = TheManagerConfig(
            steps=1,
            start_time="2022-01-01",
            seed_flag="geojson",
            geojson=geojson,
            lon=50,
            lat=50,
        )

    m = TheManagerConfig(
        steps=1,
        start_time="2022-01-01",
        seed_flag="geojson",
        geojson=geojson,
        lon=None,
        lat=None,
    )


def test_misc_parameters():
    """Test values of parameters being input."""

    m = TheManagerConfig(
        steps=1,
        start_time="2022-01-01",
        horizontal_diffusivity=1,
        number=100,
        time_step=50,
        stokes_drift=False,
    )

    assert m.horizontal_diffusivity == 1
    assert m.number == 100
    assert m.time_step == 50
    assert m.stokes_drift == False


def test_ocean_model_not_known():
    with pytest.raises(ValidationError):
        TheManagerConfig(ocean_model="wrong_name", steps=1)



"""Defines LoggerConfig to set up and manage logs."""

# Standard library imports
import logging

# Third-party imports
from pydantic import BaseModel, Field

# Local imports
from .config_the_manager import LogLevelEnum, TheManagerConfig


logger = logging.getLogger()


class LoggerConfig(BaseModel):
    """Methods for loggers."""

    log_level: LogLevelEnum = Field(
        TheManagerConfig.model_json_schema()["properties"]["log_level"]["default"]
    )

    def close_loggers(self, logger: logging.Logger) -> None:
        """Close and remove all handlers from the logger."""
        for handler in logger.handlers[:]:
            handler.close()
            logger.removeHandler(handler)

    def setup_logger(self, logfile_name: str) -> logging.Logger:
        """Setup logger."""

        logger = logging.getLogger()

        if logger.handlers:
            self.close_loggers(logger)

        logger.setLevel(getattr(logging, self.log_level))

        # Add handlers from the main logger to the OpenDrift logger if not already added

        # Create file handler to save log to file
        file_handler = logging.FileHandler(logfile_name)
        fmt = "%(asctime)s %(levelname)-7s %(name)s.%(module)s.%(funcName)s:%(lineno)d: %(message)s"
        datefmt = "%Y-%m-%d %H:%M:%S"
        formatter = logging.Formatter(fmt, datefmt)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

        # Create stream handler
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(formatter)
        logger.addHandler(stream_handler)

        logger.info("Particle tracking manager simulation.")
        logger.info(f"Log filename: {logfile_name}")
        return logger

    def merge_with_opendrift_log(self) -> None:
        """Merge the OpenDrift logger with the main logger."""
        for logger_name in logging.root.manager.loggerDict:
            if logger_name.startswith("opendrift"):
                od_logger = logging.getLogger(logger_name)
                if od_logger.handlers:
                    self.close_loggers(od_logger)

                # Add handlers from the main logger to the OpenDrift logger
                for handler in logger.handlers:
                    od_logger.addHandler(handler)
                od_logger.setLevel(logger.level)
                od_logger.propagate = False



"""Defines ParticleTrackingState and SetupOutputFiles."""

# Standard library imports
import datetime
import logging
import pathlib

from os import PathLike
from typing import Self

# Third-party imports
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    computed_field,
    create_model,
    field_validator,
    model_validator,
)

# Local imports
from .config_the_manager import OutputFormatEnum, TheManagerConfig


logger = logging.getLogger()


class ParticleTrackingState(BaseModel):
    """Track simulation state."""

    has_run_setup: bool = False  # this may not be required for all models
    has_added_reader: bool = False
    has_run_seeding: bool = False
    has_run: bool = False


def generate_default_output_file() -> str:
    """Generate a default output file name based on the current date and time."""
    return f"output-results_{datetime.datetime.now():%Y-%m-%dT%H%M%SZ}"


class SetupOutputFiles(BaseModel):
    """Handle all changes/work on output files.

    This class runs first thing. Then logger setup.
    """

    output_file: PathLike[str] | None = Field(
        TheManagerConfig.model_json_schema()["properties"]["output_file"]["default"]
    )
    output_format: OutputFormatEnum = Field(
        TheManagerConfig.model_json_schema()["properties"]["output_format"]["default"]
    )

    model_config = {"validate_default": True}

    @field_validator("output_file", mode="after")
    def assign_output_file_if_needed(value: str | None) -> str:
        """Assign a default output file name if not provided."""
        if value is None:
            return generate_default_output_file()
        return value

    @field_validator("output_file", mode="after")
    def clean_output_file(value: str) -> str:
        """Clean the output file name by removing extensions."""
        value = (
            str(value).replace(".nc", "").replace(".parquet", "").replace(".parq", "")
        )
        return value

    @model_validator(mode="after")
    def add_output_file_extension(self) -> Self:
        """Add the appropriate file extension based on the output format."""
        assert self.output_file is not None
        if self.output_format is not None:
            if self.output_format == "netcdf":
                self.output_file = pathlib.Path(self.output_file).with_suffix(".nc")
            elif self.output_format == "parquet":
                self.output_file = pathlib.Path(self.output_file).with_suffix(
                    ".parquet"
                )
            else:
                raise ValueError(f"output_format {self.output_format} not recognized.")
        return self

    @computed_field
    def logfile_name(self) -> str:
        """Generate a log file name based on the output file name."""
        assert self.output_file is not None
        return pathlib.Path(self.output_file).stem + ".log"



from pathlib import Path

import pytest

from pydantic import ValidationError

from particle_tracking_manager.models.opendrift.config_opendrift import (
    HarmfulAlgalBloomModelConfig,
    LarvalFishModelConfig,
    LeewayModelConfig,
    OceanDriftModelConfig,
    OpenDriftConfig,
    OpenOilModelConfig,
)
from particle_tracking_manager.models.opendrift.enums import ObjectTypeEnum
from particle_tracking_manager.models.opendrift.enums.species_types import (
    SPECIES_HAB_DEFAULTS,
)
from particle_tracking_manager.models.opendrift.opendrift import OpenDriftModel


def test_drift_model():
    # this test could be done on any of the drift model classes with the same result
    # i.e. LarvalFishModelConfig, LeewayModelConfig, OpenOilModelConfig, OceanDriftModelConfig
    with pytest.raises(ValidationError):
        m = OpenDriftConfig(drift_model="not_a_real_model")


## LarvalFish ##


def test_LarvalFish_init():
    m = LarvalFishModelConfig(
        drift_model="LarvalFish",
        do3D=True,
        vertical_mixing=True,
        wind_drift_factor=0,
        wind_drift_depth=0,
        steps=1,
        length=10,
    )


def test_LarvalFish_parameters():
    """Make sure LarvalFish-specific parameters are present."""
    m = LarvalFishModelConfig(drift_model="LarvalFish", steps=1)
    params = [
        "diameter",
        "neutral_buoyancy_salinity",
        "stage_fraction",
        "hatched",
        "length",
        "weight",
    ]
    for param in params:
        assert hasattr(m, param)


def test_LarvalFish_disallowed_settings():
    """LarvalFish is incompatible with some settings.

    LarvalFish has to always be 3D with vertical_mixing on.
    """

    with pytest.raises(ValidationError):
        m = LarvalFishModelConfig(
            drift_model="LarvalFish", vertical_mixing=False, steps=1
        )

    with pytest.raises(ValidationError):
        m = LarvalFishModelConfig(drift_model="LarvalFish", do3D=False, steps=1)


def test_LarvalFish_hatched_stage_fraction():
    """If hatched==1, stage_fraction must be None."""

    with pytest.raises(ValidationError):
        m = LarvalFishModelConfig(
            drift_model="LarvalFish", steps=1, hatched=1, stage_fraction=0.5
        )

    m = LarvalFishModelConfig(
        drift_model="LarvalFish", steps=1, hatched=1, stage_fraction=None
    )

    m = LarvalFishModelConfig(
        drift_model="LarvalFish", steps=1, hatched=0, stage_fraction=0.5
    )


## Leeway ##


def test_Leeway_init():
    m = LeewayModelConfig(
        drift_model="Leeway",
        do3D=False,
        steps=1,
    )


def test_Leeway_parameters():
    """Make sure Leeway-specific parameters are present."""
    m = LeewayModelConfig(drift_model="Leeway", steps=1)
    params = ["object_type"]
    for param in params:
        assert hasattr(m, param)


def test_Leeway_disallowed_settings():
    """Leeway is incompatible with some settings.

    Leeway can't have stokes drift or wind drift factor/depth or be 3D
    """

    with pytest.raises(ValidationError):
        m = LeewayModelConfig(drift_model="Leeway", stokes_drift=True, steps=1)

    with pytest.raises(ValidationError):
        m = LeewayModelConfig(
            drift_model="Leeway", wind_drift_factor=10, wind_drift_depth=10, steps=1
        )

    with pytest.raises(ValidationError):
        m = LeewayModelConfig(drift_model="Leeway", do3D=True, steps=1)


@pytest.mark.slow
def test_object_type_list():
    """Make sure options are exactly the same as in OpenDrift."""

    m = OpenDriftModel(drift_model="Leeway", steps=1)
    m.setup_for_simulation()
    od_objects = m.o.get_configspec("seed:object_type")["seed:object_type"]["enum"]

    ptm_objects = [v.value for v in ObjectTypeEnum.__members__.values()]

    assert od_objects == ptm_objects


## OceanDrift ##


def test_OceanDrift_init():
    m = OceanDriftModelConfig(
        drift_model="OceanDrift",
        steps=1,
    )


def test_OceanDrift_parameters():
    """Make sure OceanDrift-specific parameters are present."""
    m = OceanDriftModelConfig(drift_model="OceanDrift", steps=1)
    params = [
        "seed_seafloor",
        "diffusivitymodel",
        "mixed_layer_depth",
        "seafloor_action",
        "wind_drift_depth",
        "vertical_mixing_timestep",
        "wind_drift_factor",
        "vertical_mixing",
    ]
    for param in params:
        assert hasattr(m, param)


## OpenOil ##


def test_OpenOil_init():
    m = OpenOilModelConfig(
        drift_model="OpenOil",
        do3D=False,
        steps=1,
    )


def test_OpenOil_parameters():
    """Make sure OpenOil-specific parameters are present."""
    m = OpenOilModelConfig(drift_model="OpenOil", steps=1)
    params = [
        "oil_type",
        "m3_per_hour",
        "oil_film_thickness",
        "droplet_size_distribution",
        "droplet_diameter_mu",
        "droplet_diameter_sigma",
        "droplet_diameter_min_subsea",
        "droplet_diameter_max_subsea",
        "emulsification",
        "dispersion",
        "evaporation",
        "update_oilfilm_thickness",
        "biodegradation",
    ]
    for param in params:
        assert hasattr(m, param)


def test_OpenOil_json_schema():
    schema = OpenOilModelConfig.model_json_schema()
    assert "{'const': 'AD00010', 'title': 'ABU SAFAH, ARAMCO'}" in map(
        str, schema["properties"]["oil_type"]["oneOf"]
    )


## OceanDrift
def test_unknown_parameter():
    """Make sure unknown parameters are not input."""

    with pytest.raises(ValidationError):
        m = OpenDriftConfig(unknown="test", steps=1, start_time="2022-01-01")


def test_do3D_OceanDrift():
    m = OceanDriftModelConfig(
        steps=1, do3D=True, start_time="2022-01-01", vertical_mixing=True
    )
    m = OceanDriftModelConfig(
        steps=1, do3D=True, start_time="2022-01-01", vertical_mixing=False
    )
    with pytest.raises(ValidationError):
        m = OpenDriftConfig(
            steps=1, do3D=False, start_time="2022-01-01", vertical_mixing=True
        )


def test_z_OceanDrift():
    m = OceanDriftModelConfig(steps=1, start_time="2022-01-01", z=-10)
    assert m.seed_seafloor == False

    m = OceanDriftModelConfig(
        steps=1, start_time="2022-01-01", z=None, seed_seafloor=True
    )

    with pytest.raises(ValidationError):
        m = OceanDriftModelConfig(
            steps=1, start_time="2022-01-01", z=None, seed_seafloor=False
        )

    with pytest.raises(ValidationError):
        m = OceanDriftModelConfig(
            steps=1, start_time="2022-01-01", z=-10, seed_seafloor=True
        )


def test_interpolator_filename():
    with pytest.raises(ValidationError):
        m = OpenDriftConfig(interpolator_filename="test", steps=1, use_cache=False)

    m = OpenDriftConfig(interpolator_filename=None, use_cache=False, steps=1)

    m = OpenDriftConfig(use_cache=True, interpolator_filename="test", steps=1)
    assert m.interpolator_filename == Path("test.pickle")

    m = OpenDriftConfig(use_cache=True, interpolator_filename=None, steps=1)
    assert m.interpolator_filename.name == "CIOFSOP_interpolator.pickle"


## HarmfulAlgalBloom ##


def test_HarmfulAlgalBloom_init():
    m = HarmfulAlgalBloomModelConfig(
        drift_model="HarmfulAlgalBloom",
        steps=1,
    )


def test_HarmfulAlgalBloom_parameters():
    """Make sure HarmfulAlgalBloom-specific parameters are present."""
    m = HarmfulAlgalBloomModelConfig(drift_model="HarmfulAlgalBloom", steps=1)
    params = [
        "species_type",
        "temperature_death_min",
        "temperature_death_max",
        "salinity_death_min",
        "salinity_death_max",
    ]
    for param in params:
        assert hasattr(m, param)


def test_HarmfulAlgalBloom_species_type():
    """Make sure species_type parameter works as expected."""
    m = HarmfulAlgalBloomModelConfig(
        drift_model="HarmfulAlgalBloom",
        species_type="Pseudo nitzschia",
        steps=1,
    )
    assert m.species_type == "Pseudo nitzschia"
    assert (
        m.temperature_death_min
        == SPECIES_HAB_DEFAULTS["Pseudo nitzschia"].temperature_death_min
    )
    assert (
        m.temperature_death_max
        == SPECIES_HAB_DEFAULTS["Pseudo nitzschia"].temperature_death_max
    )
    assert (
        m.salinity_death_min
        == SPECIES_HAB_DEFAULTS["Pseudo nitzschia"].salinity_death_min
    )
    assert (
        m.salinity_death_max
        == SPECIES_HAB_DEFAULTS["Pseudo nitzschia"].salinity_death_max
    )


def test_HarmfulAlgalBloom_disallowed_settings():
    """HarmfulAlgalBloom is incompatible with some settings depending on the species.

    HarmfulAlgalBloom has to always be 3D with vertical_mixing on.
    """

    with pytest.raises(ValidationError):
        m = HarmfulAlgalBloomModelConfig(
            drift_model="HarmfulAlgalBloom",
            steps=1,
            species_type="custom",
        )

    # with pytest.raises(ValidationError):
    #     m = HarmfulAlgalBloomModelConfig(drift_model="HarmfulAlgalBloom", do3D=True, steps=1, species_type="Pseudo_nitzschia")



# following https://fsspec.github.io/kerchunk/test_example.html#
# env xp-inventory on lakes02
# also installed ipython, h5py, cf_xarray, pynco
# copied to /mnt/vault/ciofs/HINDCAST/ciofs_kerchunk_to2012.json

# Standard library imports
from datetime import datetime
from pathlib import Path
from typing import Callable, Iterable, List

# Third-party imports
import fsspec
import pandas as pd
import xarray as xr

from kerchunk.combine import MultiZarrToZarr


def narrow_dataset_to_simulation_time(
    ds: xr.Dataset, start_time: datetime, end_time: datetime
) -> xr.Dataset:
    """Narrow the dataset to the simulation time."""
    try:
        units = ds.ocean_time.attrs["units"]
    except KeyError:
        units = ds.ocean_time.encoding["units"]
    datestr = units.split("since ")[1]
    units_date = pd.Timestamp(datestr)

    # narrow model output to simulation time if possible before sending to Reader
    if start_time is not None and end_time is not None:
        dt_model = float(
            ds.ocean_time[1] - ds.ocean_time[0]
        )  # time step of the model output in seconds
        # want to include the next ocean model output before the first drifter simulation time
        # in case it starts before model times
        start_time_num = (start_time - units_date).total_seconds()
        if start_time_num > dt_model:
            start_time_num -= dt_model
        # want to include the next ocean model output after the last drifter simulation time
        end_time_num = (end_time - units_date).total_seconds() + dt_model
        ds = ds.sel(ocean_time=slice(start_time_num, end_time_num))

        if len(ds.ocean_time) == 0:
            raise ValueError(
                "No model output left for simulation time. Check start_time and end_time."
            )
        if len(ds.ocean_time) == 1:
            raise ValueError(
                "Only 1 model output left for simulation time. Check start_time and end_time."
            )
    else:
        raise ValueError(
            "start_time and end_time must be set to narrow model output to simulation time"
        )
    return ds


def apply_known_ocean_model_specific_changes(
    ds: xr.Dataset, ocean_model: str, use_static_masks: bool
) -> xr.Dataset:
    """Apply ocean model specific changes to the dataset.

    This includes renaming variables, adding variables, etc.
    """
    # For NWGOA, need to calculate wetdry mask from a variable
    if ocean_model == "NWGOA" and not use_static_masks:
        ds["wetdry_mask_rho"] = (~ds.zeta.isnull()).astype(int)

    # For CIOFSOP need to rename u/v to have "East" and "North" in the variable names
    # so they aren't rotated in the ROMS reader (the standard names have to be x/y not east/north)
    elif ocean_model == "CIOFSOP":
        ds = ds.rename_vars({"urot": "u_eastward", "vrot": "v_northward"})
    return ds


def apply_user_input_ocean_model_specific_changes(
    ds: xr.Dataset, use_static_masks: bool
) -> xr.Dataset:
    """Apply user input ocean model specific changes to the dataset.

    This includes renaming variables, adding variables, etc.

    For now, assume user has dropped variables ahead of time.
    """

    # check for case that self.config.use_static_masks False (which is the default)
    # but user input doesn't have wetdry masks
    # then raise exception and tell user to set use_static_masks True
    if "wetdry_mask_rho" not in ds.data_vars and not use_static_masks:
        raise ValueError(
            "User input does not have wetdry_mask_rho variable. Set use_static_masks True to use static masks instead."
        )

    # ds = ds.drop_vars(self.config.drop_vars, errors="ignore")

    return ds


def find_json_files_in_date_range(
    fs2,
    make_glob_from_year: Callable[[str], str],
    start: datetime,
    end: datetime,
    filename_date_format: str,
) -> List[str]:
    """Find JSON files in a date range and return their paths."""

    # only glob start and end year files, order isn't important
    if abs(start.year - end.year) > 1:
        raise ValueError(
            f"Start ({start.year}) and end ({end.year}) "
            "dates must be at most 1 year apart."
        )
    start_year = start.strftime("%Y")
    end_year = end.strftime("%Y")
    json_list = fs2.glob(make_glob_from_year(start_year))
    if end_year != start_year:
        json_list += fs2.glob(make_glob_from_year(end_year))

    # if going backward in time, swap start and end
    date_range_from, date_range_to = (start, end) if start <= end else (end, start)

    return [
        pth
        for pth in json_list
        if (
            date_range_from
            <= datetime.strptime(Path(pth).stem, filename_date_format)
            <= date_range_to
        )
    ]


def make_ciofs_kerchunk(start: datetime, end: datetime, name: str) -> dict:
    """_summary_

    Parameters
    ----------
    start, end : datetime
        Start and end time of the simulation.

    Returns
    -------
    kerchunk output
        _description_
    """

    if name == "CIOFS":
        output_dir_single_files = "/mnt/vault/ciofs/HINDCAST/.kerchunk_json"
    elif name == "CIOFS3":
        output_dir_single_files = "/mnt/vault/ciofs/CIOFS3/HINDCAST/.kerchunk_json"
    elif name == "CIOFSFRESH":
        output_dir_single_files = "/mnt/vault/ciofs/HINDCAST_FRESHWATER/.kerchunk_json"
    elif name == "CIOFSOP":
        output_dir_single_files = "/mnt/depot/data/packrat/prod/noaa/coops/ofs/aws_ciofs/processed/.kerchunk_json"
    else:
        raise ValueError(f"Name {name} not recognized")

    fs2 = fsspec.filesystem("")  # local file system to save final jsons to

    if name == "CIOFSOP":
        # base for matching
        def base_str(a_time: str) -> str:
            return f"{output_dir_single_files}/ciofs_{a_time}*.json"

        date_format = "ciofs_%Y-%m-%d"

    else:  # name is "CIOFS" or "CIOFSFRESH" or "CIOFS3"
        # base for matching
        def base_str(a_time: str) -> str:
            return f"{output_dir_single_files}/{a_time}_*.json"

        date_format = "%Y_0%j"

    json_list = find_json_files_in_date_range(fs2, base_str, start, end, date_format)

    # Multi-file JSONs
    # This code uses the output generated above to create a single ensemble dataset,
    # with one set of references pointing to all of the chunks in the individual files.
    # `coo_map = {"ocean_time": "cf:ocean_time"}` is necessary so that both the time
    # values and units are read and interpreted instead of just the values.

    def fix_fill_values(out: dict) -> dict:
        """Fix problem when fill_value and scara both equal 0.0.

        If the fill value and the scalar value are both 0, nan is filled instead. This fixes that.
        """

        for k in list(out):
            if isinstance(out[k], str) and '"fill_value":0.0' in out[k]:
                out[k] = out[k].replace('"fill_value":0.0', '"fill_value":"NaN"')
        return out

    def postprocess(out: dict) -> dict:
        """postprocess function to fix fill values"""
        out = fix_fill_values(out)
        return out

    mzz = MultiZarrToZarr(
        json_list,
        concat_dims=["ocean_time"],
        identical_dims=[
            "lat_rho",
            "lon_rho",
            "lon_psi",
            "lat_psi",
            "lat_u",
            "lon_u",
            "lat_v",
            "lon_v",
            "Akk_bak",
            "Akp_bak",
            "Akt_bak",
            "Akv_bak",
            "Cs_r",
            "Cs_w",
            "FSobc_in",
            "FSobc_out",
            "Falpha",
            "Fbeta",
            "Fgamma",
            "Lm2CLM",
            "Lm3CLM",
            "LnudgeM2CLM",
            "LnudgeM3CLM",
            "LnudgeTCLM",
            "LsshCLM",
            "LtracerCLM",
            "LtracerSrc",
            "LuvSrc",
            "LwSrc",
            "M2nudg",
            "M2obc_in",
            "M2obc_out",
            "M3nudg",
            "M3obc_in",
            "M3obc_out",
            "Tcline",
            "Tnudg",
            "Tobc_in",
            "Tobc_out",
            "Vstretching",
            "Vtransform",
            "Znudg",
            "Zob",
            "Zos",
            "angle",
            "dstart",
            "dt",
            "dtfast",
            "el",
            "f",
            "gamma2",
            "grid",
            "h",
            "hc",
            "mask_psi",
            "mask_rho",
            "mask_u",
            "mask_v",
            "nHIS",
            "nRST",
            "nSTA",
            "ndefHIS",
            "ndtfast",
            "ntimes",
            "pm",
            "pn",
            "rdrg",
            "rdrg2",
            "rho0",
            "spherical",
            "theta_b",
            "theta_s",
            "xl",
        ],
        coo_map={
            "ocean_time": "cf:ocean_time",
        },
        postprocess=postprocess,
    )

    # to keep in memory
    out = mzz.translate()

    return out


def make_nwgoa_kerchunk(start: datetime, end: datetime, name: str = "NWGOA") -> dict:
    """_summary_

    Parameters
    ----------
    start, end : datetime
        Start and end time of the simulation.

    Returns
    -------
    kerchunk output
        _description_
    """

    # this version of the daily json files has the grid file merged
    output_dir_single_files = (
        "/mnt/depot/data/packrat/prod/aoos/nwgoa/processed/.kerchunk_json"
    )

    fs2 = fsspec.filesystem("")  # local file system to save final jsons to

    # base for matching
    def base_str(a_time: str) -> str:
        # this is the base string for the json files
        return f"{output_dir_single_files}/nwgoa_{a_time}-*.json"

    date_format = "nwgoa_%Y-%m-%d"

    json_list = find_json_files_in_date_range(fs2, base_str, start, end, date_format)

    # account for double compression
    # Look at individual variables in the files to see what needs to be changed with
    # h5dump -d ocean_time -p /mnt/depot/data/packrat/prod/aoos/nwgoa/processed/1999/nwgoa_1999-02-01.nc
    def preprocess(refs: dict) -> dict:
        """preprocess function to fix fill values"""
        for k in list(refs):
            if k.endswith("/.zarray"):
                refs[k] = refs[k].replace(
                    '"filters":[{"elementsize":8,"id":"shuffle"}]',
                    '"filters":[{"elementsize":8,"id":"shuffle"},{"id": "zlib", "level":8}]',
                )
                refs[k] = refs[k].replace(
                    '"filters":[{"elementsize":4,"id":"shuffle"}]',
                    '"filters":[{"elementsize":4,"id":"shuffle"},{"id": "zlib", "level":8}]',
                )
        return refs

    import zarr

    def add_time_attr(out: dict) -> dict:
        """add time attributes to the ocean_time variable"""
        out_ = zarr.open(out)
        out_.ocean_time.attrs["axis"] = "T"
        return out

    def postprocess(out: dict) -> dict:
        """postprocess function to fix fill values"""
        out = add_time_attr(out)
        return out

    mzz = MultiZarrToZarr(
        json_list,
        concat_dims=["ocean_time"],
        identical_dims=[
            "lat_rho",
            "lon_rho",
            "lon_psi",
            "lat_psi",
            "lat_u",
            "lon_u",
            "lat_v",
            "lon_v",
            "Akk_bak",
            "Akp_bak",
            "Akt_bak",
            "Akv_bak",
            "Cs_r",
            "Cs_w",
            "FSobc_in",
            "FSobc_out",
            "Falpha",
            "Fbeta",
            "Fgamma",
            "Lm2CLM",
            "Lm3CLM",
            "LnudgeM2CLM",
            "LnudgeM3CLM",
            "LnudgeTCLM",
            "LsshCLM",
            "LtracerCLM",
            "LtracerSrc",
            "LuvSrc",
            "LwSrc",
            "M2nudg",
            "M2obc_in",
            "M2obc_out",
            "M3nudg",
            "M3obc_in",
            "M3obc_out",
            "Tcline",
            "Tnudg",
            "Tobc_in",
            "Tobc_out",
            "Vstretching",
            "Vtransform",
            "Znudg",
            "Zob",
            "Zos",
            "angle",
            "dstart",
            "dt",
            "dtfast",
            "el",
            "f",
            "gamma2",
            "grid",
            "h",
            "hc",
            "mask_psi",
            "mask_rho",
            "mask_u",
            "mask_v",
            "nHIS",
            "nRST",
            "nSTA",
            "ndefHIS",
            "ndtfast",
            "ntimes",
            "pm",
            "pn",
            "rdrg",
            "rdrg2",
            "rho0",
            "spherical",
            "theta_b",
            "theta_s",
            "xl",
            "Charnok_alpha",
            "CrgBan_cw",
            "JLTS",
            "JPRJ",
            "LuvSponge",
            "P1",
            "P2",
            "P3",
            "P4",
            "PLAT",
            "PLONG",
            "ROTA",
            "XOFF",
            "YOFF",
            "Zos_hsig_alpha",
            "depthmax",
            "depthmin",
            "dfdy",
            "dmde",
            "dndx",
            "f0",
            "gls_Kmin",
            "gls_Pmin",
            "gls_c1",
            "gls_c2",
            "gls_c3m",
            "gls_c3p",
            "gls_cmu0",
            "gls_m",
            "gls_n",
            "gls_p",
            "gls_sigk",
            "gls_sigp",
            "h_mask",
            "hraw",
            "nAVG",
            "ndefAVG",
            "nl_visc2",
            "ntsAVG",
            "sz_alpha",
            "wtype_grid",
            "x_psi",
            "x_rho",
            "x_u",
            "x_v",
            "y_psi",
            "y_rho",
            "y_u",
            "y_v",
        ],
        coo_map={
            "ocean_time": "cf:ocean_time",
            #    "eta_rho": list(np.arange(1044))
        },
        preprocess=preprocess,
        postprocess=postprocess,
    )

    # to keep in memory
    out = mzz.translate()

    return out



"""Some extra plotting code."""

import opendrift


def plot_dest(o, filename=None, **kwargs):
    """This is copied from an opendrift example.

    ...and subsequently modified.

    Parameters
    ----------
    kwargs : dict
        Additional keyword arguments to pass to the plotting function.
    """

    import cmocean

    cmap = cmocean.tools.crop_by_percent(cmocean.cm.amp, 20, which="max", N=None)

    od = opendrift.open_xarray(o.outfile_name)

    density = od.get_histogram(pixelsize_m=5000).isel(time=-1).isel(origin_marker=0)
    density = density.where(density > 0)
    density = density / density.sum() * 100
    vmax = density.max()

    return o.plot(
        background=density,
        clabel="Probability of final location [%]",
        markersize=0.5,
        lalpha=0.02,
        vmin=0,
        vmax=vmax,
        cmap=cmap,
        fast=True,
        filename=filename,
        **kwargs
    )



from particle_tracking_manager.config_misc import SetupOutputFiles


def test_log_name():
    m = SetupOutputFiles(output_file="newtest")
    assert m.logfile_name == "newtest.log"

    m = SetupOutputFiles(output_file="newtest.nc")
    assert m.logfile_name == "newtest.log"

    m = SetupOutputFiles(output_file="newtest.parq")
    assert m.logfile_name == "newtest.log"

    m = SetupOutputFiles(output_file="newtest.parquet")
    assert m.logfile_name == "newtest.log"


def test_output_file():
    """make sure output file is parquet if output_format is parquet"""

    m = SetupOutputFiles(output_format="parquet")
    assert m.output_file.suffix == ".parquet"

    m = SetupOutputFiles(output_format="netcdf")
    assert m.output_file.suffix == ".nc"



"""Particle Tracking Manager."""

import logging


# Set log levels for third paries to WARNING
logging.getLogger("fsspec").setLevel(logging.WARNING)
logging.getLogger("kerchunk").setLevel(logging.WARNING)
logging.getLogger("opendrift").setLevel(logging.WARNING)
logging.getLogger("numcodecs").setLevel(logging.WARNING)

from .config_the_manager import TheManagerConfig
from .models import opendrift as opendrift_models
from .models.opendrift.config_opendrift import (
    HarmfulAlgalBloomModelConfig,
    LarvalFishModelConfig,
    LeewayModelConfig,
    OceanDriftModelConfig,
    OpenDriftConfig,
    OpenOilModelConfig,
)
from .models.opendrift.opendrift import OpenDriftModel


__all__ = [
    "TheManagerConfig",
    "LarvalFishModelConfig",
    "LeewayModelConfig",
    "OceanDriftModelConfig",
    "OpenDriftConfig",
    "OpenOilModelConfig",
    "OpenDriftModel",
    "HarmfulAlgalBloomModelConfig",
    "opendrift_models",
]



"""Test CLI methods."""

import os

from datetime import datetime

import pytest

import particle_tracking_manager as ptm


@pytest.mark.slow
def test_setup():
    """Test CLI setup with dryrun

    No drifters are run due to dryrun flag
    """
    ret_value = os.system(f"ptm steps=1 --dry-run")
    assert ret_value == 0


def test_setup_library():
    """Same test but with library"""

    m = ptm.OpenDriftModel(
        steps=3,
    )
    m.config.model_dump()



"""OpenDrift HAB species types enum definition."""

from enum import Enum

from pydantic import BaseModel, ConfigDict, Field


class HABSpeciesTypeEnum(str, Enum):
    """Enum for species types used in HarmfulAlgalBloom model in OpenDrift."""

    PN = "Pseudo nitzschia"
    custom = "custom"  # for fully user-specified params


class HABParameters(BaseModel):
    model_config = ConfigDict(extra="forbid")

    temperature_death_min: float = Field(
        description="Minimum temperature for living. Below this temperature, cells have high mortality rate. Between this and temperature_viable_min, cells have no growth.",
        title="Cell death below this temperature",
        ge=-5.0,
        le=40.0,
        default=3.0,
        json_schema_extra={
            "units": "degrees",
            "od_mapping": "hab:temperature_death_min",
            "ptm_level": 2,
        },
    )

    temperature_death_max: float = Field(
        description="Maximum temperature for living. Above this temperature, cells have high mortality rate. Between temperature_viable_max and this parameter, cells have no growth.",
        title="Cell death above this temperature",
        ge=-5.0,
        le=40.0,
        default=22.0,
        json_schema_extra={
            "units": "degrees",
            "od_mapping": "hab:temperature_death_max",
            "ptm_level": 2,
        },
    )

    mortality_rate_high: float = Field(
        description="Rate of mortality applied below temperature_death_min and above temperature_death_max.",
        title="High mortality rate",
        ge=0.0,
        le=10.0,
        default=1.0,
        json_schema_extra={
            "units": "days^-1",
            "od_mapping": "hab:mortality_rate_high",
            "ptm_level": 2,
        },
    )

    salinity_death_min: float = Field(
        description="Minimum salinity for living. Below this salinity, cells have high mortality rate. Between this and salinity_viable_min, cells have no growth.",
        title="Cell death below this salinity",
        ge=0.0,
        le=50.0,
        default=25.0,
        json_schema_extra={
            "units": "psu",
            "od_mapping": "hab:salinity_death_min",
            "ptm_level": 2,
        },
    )

    salinity_death_max: float = Field(
        description="Maximum salinity for living. Above this salinity, cells have high mortality rate. Between salinity_viable_max and this parameter, cells have no growth.",
        title="Cell death above this salinity",
        ge=0.0,
        le=50.0,
        default=36.0,
        json_schema_extra={
            "units": "psu",
            "od_mapping": "hab:salinity_death_max",
            "ptm_level": 2,
        },
    )


# Species default for HarmfulAlgalBloom model
SPECIES_HAB_DEFAULTS: dict[HABSpeciesTypeEnum, HABParameters] = {
    HABSpeciesTypeEnum.PN: HABParameters(
        temperature_death_min=3.0,
        temperature_death_max=22.0,
        mortality_rate_high=1.0,
        salinity_death_min=25.0,
        salinity_death_max=36.0,
    ),
    # HABSpeciesTypeEnum.custom intentionally has no entry
}


# Other config defaults per species (z, do3D, etc.)
SPECIES_HAB_MANAGER_DEFAULTS: dict[HABSpeciesTypeEnum, dict[str, object]] = {
    HABSpeciesTypeEnum.PN: {
        "z": -1.0,
        "do3D": False,
    },
    # HABSpeciesTypeEnum.Alexandrium: {"do3D": True, ...},
    # HABSpeciesTypeEnum.Dinophysis: {...},
}

# this is for the schema
_species_descriptions = {
    species.value: "Defaults: "
    + ", ".join(
        [
            f"{key}={value}"
            for key, value in SPECIES_HAB_DEFAULTS[species].model_dump().items()
        ]
    )
    for species in HABSpeciesTypeEnum
    if species != "custom"
}
_species_descriptions["custom"] = "Custom species with user-defined parameters."



"""OpenDrift object types enum definition.

How to set up the enum:
import numpy as np
import particle_tracking_manager as ptm
m = ptm.OpenDriftModel(drift_model="Leeway", steps=1)
m.setup_for_simulation()
objects = m.o.get_configspec("seed:object_type")["seed:object_type"]["enum"]
nums = np.arange(0, len(objects))
for i, s in zip(nums, objects):
    print(f'    O{i+1} = "{s}"')
"""

from enum import Enum


class ObjectTypeEnum(str, Enum):
    """Enum for object types used in OpenDrift.

    From https://github.com/OpenDrift/opendrift/blob/master/opendrift/models/OBJECTPROP.DAT
    """

    O1 = "Person-in-water (PIW), unknown state (mean values)"
    O2 = ">PIW, vertical PFD type III conscious"
    O3 = ">PIW, sitting, PFD type I or II"
    O4 = ">PIW, survival suit (face up)"
    O5 = ">PIW, scuba suit (face up)"
    O6 = ">PIW, deceased (face down)"
    O7 = "Life raft, deep ballast (DB) system, general, unknown capacity and loading (mean values)"
    O8 = ">4-14 person capacity, deep ballast system, canopy (average)"
    O9 = ">>4-14 person capacity, deep ballast system, no drogue"
    O10 = (
        ">>>4-14 person capacity, deep ballast system, canopy, no drogue, light loading"
    )
    O11 = ">>>4-14 person capacity, deep ballast system, no drogue, heavy loading"
    O12 = ">>4-14 person capacity, deep ballast system, canopy, with drogue (average)"
    O13 = ">>>4-14 person capacity, deep ballast system, canopy, with drogue, light loading"
    O14 = ">>>4-14 person capacity, deep ballast system, canopy, with drogue, heavy loading"
    O15 = ">15-50 person capacity, deep ballast system, canopy, general (mean values)"
    O16 = (
        ">>15-50 person capacity, deep ballast system, canopy, no drogue, light loading"
    )
    O17 = ">>15-50 person capacity, deep ballast system, canopy, with drogue, heavy loading"
    O18 = "Deep ballast system, general (mean values), capsized"
    O19 = "Deep ballast system, general (mean values), swamped"
    O20 = "Life-raft, shallow ballast (SB) system AND canopy, general (mean values)"
    O21 = ">Life-raft, shallow ballast system, canopy, no drogue"
    O22 = ">Life-raft, shallow ballast system AND canopy, with drogue"
    O23 = "Life-raft, shallow ballast system AND canopy, capsized"
    O24 = "Life Raft - Shallow ballast, canopy, Navy Sub Escape (SEIE) 1-man raft, NO drogue"
    O25 = "Life Raft - Shallow ballast, canopy, Navy Sub Escape (SEIE) 1-man raft, with drogue"
    O26 = "Life-raft, no ballast (NB) system, general (mean values)"
    O27 = ">Life-raft, no ballast system, no canopy, no drogue"
    O28 = ">Life-raft, no ballast system, no canopy, with drogue"
    O29 = ">Life-raft, no ballast system, with canopy, no drogue"
    O30 = ">Life-raft, no ballast system, with canopy, with drogue"
    O31 = "Survival Craft - USCG Sea Rescue Kit - 3 ballasted life rafts and 300 meter of line"
    O32 = "Life-raft, 4-6 person capacity, no ballast, with canopy, no drogue"
    O33 = "Evacuation slide with life-raft, 46 person capacity"
    O34 = "Survival Craft - SOLAS Hard Shell Life Capsule, 22 man"
    O35 = "Survival Craft - Ovatek Hard Shell Life Raft, 4 and 7-man, lightly loaded, no drogue (average)"
    O36 = ">Survival Craft - Ovatek Hard Shell Life Raft, 4 man, lightly loaded, no drogue"
    O37 = ">Survival Craft - Ovatek Hard Shell Life Raft, 7 man, lightly loaded, no drogue"
    O38 = "Survival Craft - Ovatek Hard Shell Life Raft, 4 and 7-man, fully loaded, drogued (average)"
    O39 = ">Survival Craft - Ovatek Hard Shell Life Raft, 4 man, fully loaded, drogued"
    O40 = ">Survival Craft - Ovatek Hard Shell Life Raft, 7 man, fully loaded, drogued"
    O41 = "Sea Kayak with person on aft deck"
    O42 = "Surf board with person"
    O43 = "Windsurfer with mast and sail in water"
    O44 = "Skiff - modified-v, cathedral-hull, runabout outboard powerboat"
    O45 = "Skiff, V-hull"
    O46 = "Skiffs, swamped and capsized"
    O47 = "Skiff - v-hull bow to stern (aluminum, Norway)"
    O48 = "Sport boat, no canvas (*1), modified V-hull"
    O49 = "Sport fisher, center console (*2), open cockpit"
    O50 = "Fishing vessel, general (mean values)"
    O51 = "Fishing vessel, Hawaiian Sampan (*3)"
    O52 = ">Fishing vessel, Japanese side-stern trawler"
    O53 = ">Fishing vessel, Japanese Longliner (*3)"
    O54 = ">Fishing vessel, Korean fishing vessel (*4)"
    O55 = ">Fishing vessel, Gill-netter with rear reel (*3)"
    O56 = "Coastal freighter. (*5)"
    O57 = "Sailboat Mono-hull (Average)"
    O58 = ">Sailboat Mono-hull (Dismasted, Average)"
    O59 = ">>Sailboat Mono-hull (Dismasted - rudder amidships)"
    O60 = ">>Sailboat Mono-hull (Dismasted - rudder missing)"
    O61 = ">Sailboat Mono-hull (Bare-masted,  Average)"
    O62 = ">>Sailboat Mono-hull (Bare-masted, rudder amidships)"
    O63 = ">>Sailboat Mono-hull (Bare-masted, rudder hove-to)"
    O64 = "Sailboat Mono-hull, fin keel, shallow draft (was SAILBOAT-2)"
    O65 = "Sunfish sailing dingy  -  Bare-masted, rudder missing"
    O66 = "Fishing vessel debris"
    O67 = "Self-locating datum marker buoy - no windage"
    O68 = "Navy Submarine EPIRB (SEPIRB)"
    O69 = "Bait/wharf box, holds a cubic metre of ice, mean values (*6)"
    O70 = "Bait/wharf box, holds a cubic metre of ice, lightly loaded"
    O71 = ">Bait/wharf box, holds a cubic metre of ice, full loaded"
    O72 = "55-gallon (220 l) Oil Drum"
    O73 = "Scaled down (1:3) 40-ft Container (70% submerged)"
    O74 = "20-ft Container (80% submerged)"
    O75 = "WWII L-MK2 mine"
    O76 = "Immigration vessel, Cuban refugee-raft, no sail (*7)"
    O77 = "Immigration vessel, Cuban refugee-raft, with sail (*7)"
    O78 = "Sewage floatables, tampon applicator"
    O79 = "Medical waste (mean values)"
    O80 = ">Medical waste, vials"
    O81 = ">>Medical waste, vials, large"
    O82 = ">>Medical waste, vials, small"
    O83 = ">Medical waste, syringes"
    O84 = ">>Medical waste, syringes, large"
    O85 = ">>Medical waste, syringes, small"



"""Defines TheManagerConfig class for the particle tracking manager."""

# Standard library imports
import logging

from datetime import datetime, timedelta
from enum import Enum
from os import PathLike
from typing import Annotated, Self

# Third-party imports
import pandas as pd

# from geojson import GeoJSON
from pydantic import BaseModel, Field, computed_field, model_validator

# Local imports
from .config_ocean_model import (
    OceanModelEnum,
    OceanModelSimulation,
    ocean_model_simulation_mapper,
    update_TXLA_with_download_location,
)
from .ocean_model_registry import OceanModelConfig, ocean_model_registry


logger = logging.getLogger()


# Enum for "model"
class ModelEnum(str, Enum):
    """Lagrangian model software to use for simulation."""

    opendrift = "opendrift"


# Enum for "seed_flag"
class SeedFlagEnum(str, Enum):
    """Method for seeding drifters."""

    elements = "elements"
    geojson = "geojson"


# Enum for "output_format"
class OutputFormatEnum(str, Enum):
    """Output file format."""

    netcdf = "netcdf"
    parquet = "parquet"


# Enum for "log_level"
class LogLevelEnum(str, Enum):
    """Log verbosity."""

    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


# from geojson_pydantic import LineString, Point, Polygon


class TheManagerConfig(BaseModel):
    """Configuration for the particle tracking manager."""

    model: ModelEnum = Field(
        ModelEnum.opendrift,
        description="Lagrangian model software to use for simulation.",
        json_schema_extra=dict(ptm_level=1),
    )
    lon: float | None = Field(
        -151.0,
        ge=-180,
        le=180,
        description='Central longitude for seeding drifters. Only used if `seed_flag=="elements"`.',
        json_schema_extra=dict(ptm_level=1, units="degrees_east"),
    )
    lat: float | None = Field(
        58.0,
        ge=-90,
        le=90,
        description='Central latitude for seeding drifters. Only used if `seed_flag=="elements"`.',
        json_schema_extra=dict(ptm_level=1, units="degrees_north"),
    )
    geojson: dict | None = Field(
        None,
        description='GeoJSON describing a polygon within which to seed drifters. To use this parameter, also have `seed_flag=="geojson"`.',
        json_schema_extra=dict(ptm_level=1),
    )
    # geojson: GeoJSON | None = Field(
    #     None,
    #     description='GeoJSON describing a polygon within which to seed drifters. To use this parameter, also have `seed_flag=="geojson"`.',
    #     json_schema_extra=dict(ptm_level=1),
    # )
    # geojson: Annotated[
    #     Point | LineString | Polygon | None,
    #     Field(
    #         None,
    #         description="GeoJSON describing a point, line, or polygon for seeding drifters.",  # noqa: E501
    #     ),
    # ]
    seed_flag: SeedFlagEnum = Field(
        SeedFlagEnum.elements,
        description='Method for seeding drifters. Options are "elements" or "geojson". If "elements", seed drifters at or around a single point defined by lon and lat. If "geojson", seed drifters within a polygon described by a GeoJSON object.',
        json_schema_extra=dict(ptm_level=1),
    )
    start_time: datetime | None = Field(
        datetime(2022, 1, 1),
        description="Start time for drifter simulation. start_time or end_time must be input. If a timezone is included, it will be used and the time will be converted to UTC which is the same timezone as the models.",
        json_schema_extra=dict(ptm_level=1),
    )
    start_time_end: datetime | None = Field(
        None,
        description="If used, this creates a range of start times for drifters, starting with `start_time` and ending with `start_time_end`. Drifters will be initialized linearly between the two start times. If a timezone is included, it will be used and the time will be converted to UTC which is the same timezone as the models.",
        json_schema_extra=dict(ptm_level=2),
    )
    run_forward: bool = Field(
        True, description="Run forward in time.", json_schema_extra=dict(ptm_level=2)
    )
    time_step: float = Field(
        300,
        ge=0.01,
        le=1440,
        description="Interval between particles updates, in seconds.",
        json_schema_extra=dict(ptm_level=3, units="seconds"),
    )
    time_step_output: float = Field(
        3600,
        ge=1,
        le=86400,
        description="Time step at which element properties are stored and eventually written to file. This must be larger than the calculation time step, and be an integer multiple of this.",
        json_schema_extra=dict(ptm_level=3, units="seconds"),
    )
    steps: int | None = Field(
        None,
        ge=1,
        le=10000,
        description="Maximum number of steps. End of simulation will be start_time + steps * time_step.",
        json_schema_extra=dict(ptm_level=1),
    )
    duration: str | None = Field(
        None,
        description="Duration should be input as a string of ISO 8601. The length of the simulation. steps, end_time, or duration must be input by user.",
        json_schema_extra=dict(ptm_level=1),
    )
    end_time: datetime | None = Field(
        None,
        description="The end of the simulation. steps, end_time, or duration must be input by user. start_time or end_time must be input. If a timezone is included, it will be used and the time will be converted to UTC which is the same timezone as the models.",
        json_schema_extra=dict(ptm_level=1),
    )
    # OceanModelEnum was created dynamically and will trigger an issue with mypy, so we suppress it
    ocean_model: OceanModelEnum = Field(  # type: ignore
        "CIOFSOP",
        description="Name of ocean model to use for driving drifter simulation.",
        json_schema_extra=dict(ptm_level=1),
    )
    ocean_model_local: bool = Field(
        True,
        description="Set to True to use local version of known `ocean_model` instead of remote version.",
        json_schema_extra=dict(ptm_level=3),
    )
    do3D: bool = Field(
        False,
        description="Set to True to run drifters in 3D, by default False for most drift models.",
        json_schema_extra=dict(ptm_level=1),
    )
    use_static_masks: bool = Field(
        True,
        description="If False, use static ocean model land masks. This saves some computation time but since the available ocean models run in wet/dry mode, it is inconsistent with the ROMS output files in some places since the drifters may be allowed (due to the static mask) to enter a cell they wouldn't otherwise. However, it doesn't make much of a difference for simulations that aren't in the tidal flats. Use the time-varying wet/dry masks (set to True) if drifters are expected to run in the tidal flats. This costs some more computational time but is fully consistent with the ROMS output files.",
        json_schema_extra=dict(ptm_level=3),
    )
    output_file: PathLike[str] | None = Field(
        None,
        description="Name of file to write output to. If None, default name is used.",
        json_schema_extra=dict(ptm_level=3),
    )
    output_format: OutputFormatEnum = Field(
        OutputFormatEnum.netcdf,
        description='Output file format. Options are "netcdf" or "parquet".',
        json_schema_extra=dict(ptm_level=2),
    )
    use_cache: bool = Field(
        True,
        description="Set to True to use cache for storing interpolators.",
        json_schema_extra=dict(ptm_level=3),
    )
    log_level: LogLevelEnum = Field(
        LogLevelEnum.INFO,
        description="Log verbosity",
        json_schema_extra=dict(ptm_level=3),
    )
    # TODO: change log_level to "verbose" or similar

    horizontal_diffusivity: float | None = Field(
        default=None,
        description="Add horizontal diffusivity (random walk). For known ocean models, the value is calculated as the approximate horizontal grid resolution for the selected ocean model times an estimate of the scale of sub-gridscale velocity of 0.1 m/s.",
        title="Horizontal Diffusivity",
        ge=0,
        le=100000,
        json_schema_extra=dict(units="m2/s", ptm_level=2),
    )

    stokes_drift: bool = Field(
        default=True,
        description="Advection elements with Stokes drift (wave orbital motion).",
        title="Stokes Drift",
        json_schema_extra=dict(ptm_level=2),
    )

    z: float | None = Field(
        default=0,
        description="Depth below sea level where elements are released. This depth is neglected if seafloor seeding is set selected.",
        title="Z",
        le=0,
        ge=-10000,
        json_schema_extra=dict(units="m", ptm_level=1),
    )

    number: int = Field(
        default=1,
        description="The number of elements for the simulation.",
        title="Number",
        ge=1,
        json_schema_extra=dict(
            units=1,
            ptm_level=1,
        ),
    )

    model_config = {
        "validate_default": True,
        # Field values will be returned as the `enum.value` - a string in most our cases
        "use_enum_values": True,
        "extra": "forbid",
    }

    @model_validator(mode="after")
    def check_config_seed_flag_elements(self) -> Self:
        """Check if lon and lat are set when using seed_flag 'elements'."""
        if self.seed_flag == "elements" and (self.lon is None or self.lat is None):
            raise ValueError(
                'lon and lat need non-None values if using `seed_flag="elements"`.'
            )
        return self

    @model_validator(mode="after")
    def check_config_seed_flag_geojson(self) -> Self:
        """Check if geojson is set when using seed_flag 'geojson'."""
        if self.seed_flag == "geojson" and self.geojson is None:
            raise ValueError(
                'geojson need non-None value if using `seed_flag="geojson"`.'
            )
        if self.seed_flag == "geojson" and (
            self.lon is not None or self.lat is not None
        ):
            raise ValueError(
                'lon and lat need to be None if using `seed_flag="geojson"`.'
            )
        return self

    @model_validator(mode="after")
    def check_config_time_parameters(self) -> Self:
        """Check if exactly two of start_time, end_time, duration, and steps are set."""
        non_none_count = sum(
            x is not None
            for x in [self.start_time, self.end_time, self.duration, self.steps]
        )
        if non_none_count == 4:
            # calculate duration and steps from start_time and end_time and make sure they are the same as what
            # is already saved.
            assert self.start_time is not None
            assert self.end_time is not None
            duration = pd.Timedelta(abs(self.end_time - self.start_time)).isoformat()
            steps = int(
                abs(self.end_time - self.start_time) / timedelta(seconds=self.time_step)
            )
            if duration != self.duration:
                raise ValueError(
                    f"duration and calculated duration are inconsistent: {self.duration} != {duration}"
                )
            if steps != self.steps:
                raise ValueError(
                    f"steps and calculated steps are inconsistent: {self.steps} != {steps}"
                )
        elif non_none_count != 2:
            raise ValueError(
                f"Exactly two of start_time, end_time, duration, and steps must be non-None. "
                f"Current values are: start_time={self.start_time}, end_time={self.end_time}, "
                f"duration={self.duration}, steps={self.steps}."
            )
        if self.start_time is None and self.end_time is None:
            raise ValueError("One of start_time or end_time must be non-None.")
        return self

    @computed_field
    def timedir(self) -> int:
        """Set timedir to 1 for forward, -1 for backward."""
        if self.run_forward:
            value = 1
            logger.info("Running model forward in time.")
        else:
            value = -1
            logger.info("Running model backward in time.")
        return value

    @model_validator(mode="after")
    def match_time_step_sign_to_timedir_sign(self) -> Self:
        """If simulation is backward, make time_step negative.

        Sign of input time_step is ignored.
        """
        self.time_step = abs(self.time_step) * self.timedir
        logger.debug(f"Setting time_step to {self.time_step} to match timedir sign.")
        return self

    @model_validator(mode="after")
    def calculate_config_times(self) -> Self:
        """Calculate start_time, end_time, duration, and steps based on the other parameters."""
        if self.steps is None:
            if self.duration is not None:
                self.steps = int(
                    pd.Timedelta(self.duration) / pd.Timedelta(seconds=self.time_step)
                )
                logger.debug(f"Setting steps to {self.steps} based on duration.")
            elif self.end_time is not None and self.start_time is not None:
                self.steps = int(
                    abs(self.end_time - self.start_time)
                    / timedelta(seconds=self.time_step)
                )
                logger.debug(
                    f"Setting steps to {self.steps} based on end_time and start_time."
                )
            else:
                raise ValueError("steps has not been calculated")

        if self.duration is None:
            if self.end_time is not None and self.start_time is not None:
                self.duration = pd.Timedelta(
                    abs(self.end_time - self.start_time)
                ).isoformat()
                # # convert to ISO 8601 string
                # self.duration = pd.Timedelta(abs(self.end_time - self.start_time)).isoformat()
                logger.debug(
                    f"Setting duration to {self.duration} based on end_time and start_time."
                )
            elif self.steps is not None:
                self.duration = pd.Timedelta(
                    self.steps * timedelta(seconds=self.time_step)
                ).isoformat()
                # # convert to ISO 8601 string
                # self.duration = (self.steps * pd.Timedelta(seconds=self.time_step)).isoformat()
                logger.debug(f"Setting duration to {self.duration} based on steps.")
            else:
                raise ValueError("duration has not been calculated")

        if self.end_time is None:
            if self.steps is not None and self.start_time is not None:
                self.end_time = self.start_time + self.timedir * self.steps * timedelta(
                    seconds=self.time_step
                )
                logger.debug(
                    f"Setting end_time to {self.end_time} based on start_time and steps."
                )
            elif self.duration is not None and self.start_time is not None:
                self.end_time = self.start_time + self.timedir * self.duration
                logger.debug(
                    f"Setting end_time to {self.end_time} based on start_time and duration."
                )
            else:
                raise ValueError("end_time has not been calculated")

        if self.start_time is None:
            if self.end_time is not None and self.steps is not None:
                self.start_time = self.end_time - self.timedir * self.steps * timedelta(
                    seconds=self.time_step
                )
                logger.debug(
                    f"Setting start_time to {self.start_time} based on end_time and steps."
                )
            elif self.duration is not None and self.end_time is not None:
                self.start_time = self.end_time - self.timedir * self.duration
                logger.debug(
                    f"Setting start_time to {self.start_time} based on end_time and duration."
                )
            else:
                raise ValueError("start_time has not been calculated")

        return self

    @model_validator(mode="after")
    def remove_timezone_info(self) -> Self:
        """Remove timezone information from datetime fields."""
        if self.start_time is not None and self.start_time.tzinfo is not None:
            self.start_time = (
                pd.Timestamp(self.start_time)
                .tz_convert("UTC")
                .tz_localize(None)
                .to_pydatetime()
            )
            logger.debug(
                f"Removed timezone info from start_time, new value is {self.start_time}."
            )
        if self.start_time_end is not None and self.start_time_end.tzinfo is not None:
            self.start_time_end = (
                pd.Timestamp(self.start_time_end)
                .tz_convert("UTC")
                .tz_localize(None)
                .to_pydatetime()
            )
            logger.debug(
                f"Removed timezone info from start_time_end, new value is {self.start_time_end}."
            )
        if self.end_time is not None and self.end_time.tzinfo is not None:
            self.end_time = (
                pd.Timestamp(self.end_time)
                .tz_convert("UTC")
                .tz_localize(None)
                .to_pydatetime()
            )
            logger.debug(
                f"Removed timezone info from end_time, new value is {self.end_time}."
            )
        return self

    @computed_field
    def ocean_model_config(self) -> OceanModelConfig:
        """Select ocean model config based on ocean_model."""
        return ocean_model_registry.get(self.ocean_model)

    @computed_field
    def ocean_model_simulation(self) -> OceanModelSimulation:
        """Select ocean model simulation based on ocean_model."""
        inputs = {
            "lon": self.lon,
            "lat": self.lat,
            "start_time": self.start_time,
            "end_time": self.end_time,
            "ocean_model_local": self.ocean_model_local,
        }
        return ocean_model_simulation_mapper[self.ocean_model](**inputs)

    @model_validator(mode="after")
    def select_ocean_model_simulation_on_init(self) -> Self:
        """Select ocean model simulation based on ocean_model."""
        self.ocean_model_simulation
        return self

    @model_validator(mode="after")
    def assign_horizontal_diffusivity(self) -> Self:
        """Calculate horizontal diffusivity based on ocean model."""

        # check horizontal_diffusivity from TheManagerConfig
        if self.horizontal_diffusivity is not None:
            logger.debug(
                f"Setting horizontal_diffusivity to user-selected value {self.horizontal_diffusivity}."
            )

        # otherwise use ocean_model_config version of horizontal_diffusivity
        elif (
            self.ocean_model_config is not None
            and self.ocean_model_config.name in ocean_model_registry.all()
        ):

            self.horizontal_diffusivity = self.ocean_model_config.horizontal_diffusivity
            logger.debug(
                f"Setting horizontal_diffusivity parameter to one tuned to reader model of value {self.horizontal_diffusivity}."
            )

        elif (
            self.ocean_model_config is not None
            and self.ocean_model_config.name not in ocean_model_registry.all()
            and self.horizontal_diffusivity is None
        ):

            logger.debug(
                """Since ocean_model is user-input, changing horizontal_diffusivity parameter from None to 0.0.
                You can also set it to a specific value with `m.horizontal_diffusivity=[number]`."""
            )

            self.horizontal_diffusivity = 0

        return self

    @model_validator(mode="after")
    def check_config_ocean_model_local(self) -> Self:
        """Descrive how ocean_model_local is set."""
        if self.ocean_model_local:
            logger.debug("Using local output for ocean_model.")
        else:
            logger.debug("Using remote output for ocean_model.")
        return self

    @model_validator(mode="after")
    def setup_TXLA_if_used(self) -> Self:
        """Set up TXLA if used."""
        if self.ocean_model == "TXLA":
            update_TXLA_with_download_location()
        return self



import importlib
import os
import tempfile

from datetime import datetime
from pathlib import Path
from unittest.mock import patch

import pytest
import yaml

from pydantic import ValidationError

import particle_tracking_manager

from particle_tracking_manager.config_ocean_model import ocean_model_simulation_mapper
from particle_tracking_manager.models.opendrift.utils import (
    find_json_files_in_date_range,
)
from particle_tracking_manager.ocean_model_registry import ocean_model_registry


# Valid values
# end_time calculated as 1 5-minute step
tests_valid = {
    "NWGOA": {
        "lon": -147,
        "lat": 59,
        "start_time": "2007-01-01T00:00",
        "end_time": "2007-01-01T05:00",
    },
    "CIOFS": {
        "lon": -153,
        "lat": 59,
        "start_time": "2021-01-01T00:00",
        "end_time": "2021-01-01T05:00",
    },
    "CIOFSOP": {
        "lon": -153,
        "lat": 59,
        "start_time": "2022-01-01T00:00",
        "end_time": "2022-01-01T05:00",
    },
    "CIOFSFRESH": {
        "lon": -153,
        "lat": 59,
        "start_time": "2014-01-01T00:00",
        "end_time": "2014-01-01T05:00",
    },
    "CIOFS3": {
        "lon": -153,
        "lat": 59,
        "start_time": "2023-01-01T00:00",
        "end_time": "2023-01-01T05:00",
    },
}

# Invalid values (except start_times are valid since not testing those here)
tests_invalid = {
    "NWGOA": {
        "lon": 185 - 360,
        "lat": 50,
        "start_time": "2022-01-01",
        "end_time": "2022-01-01T05:00",
        "ocean_model_config": ocean_model_registry.get("NWGOA"),
    },
    "CIOFS": {
        "lon": -145,
        "lat": 40,
        "start_time": "2024-01-01",
        "end_time": "2024-01-01T05:00",
        "ocean_model_config": ocean_model_registry.get("CIOFS"),
    },
    "CIOFSOP": {
        "lon": -145,
        "lat": 40,
        "start_time": "2020-01-01",
        "end_time": "2020-01-01T05:00",
        "ocean_model_config": ocean_model_registry.get("CIOFSOP"),
    },
    "CIOFSFRESH": {
        "lon": -145,
        "lat": 40,
        "start_time": "2022-01-01",
        "end_time": "2022-01-01T05:00",
        "ocean_model_config": ocean_model_registry.get("CIOFSFRESH"),
    },
    "CIOFS3": {
        "lon": -145,
        "lat": 40,
        "start_time": "2026-01-01",
        "end_time": "2026-01-01T05:00",
        "ocean_model_config": ocean_model_registry.get("CIOFS3"),
    },
}


def test_lon_lat():
    """Check for valid lon and lat values

    ...for specific ocean models of type OceanModelSimulation.
    Checked for general ranges in test_config_the_manager.py.
    """

    for ocean_model, test in tests_valid.items():
        m = ocean_model_simulation_mapper[ocean_model](
            start_time=test["start_time"],
            end_time=test["end_time"],
            lon=test["lon"],
            lat=test["lat"],
            ocean_model_local=True,
        )

    for ocean_model, test in tests_invalid.items():
        with pytest.raises(ValidationError):
            m = ocean_model_simulation_mapper[ocean_model](
                start_time=tests_valid[ocean_model]["start_time"],
                end_time=tests_valid[ocean_model]["end_time"],
                lon=test["lon"],
                lat=tests_valid[ocean_model]["lat"],
                ocean_model_local=True,
            )
        with pytest.raises(ValidationError):
            m = ocean_model_simulation_mapper[ocean_model](
                start_time=tests_valid[ocean_model]["start_time"],
                end_time=tests_valid[ocean_model]["end_time"],
                lon=tests_valid[ocean_model]["lon"],
                lat=test["lat"],
                ocean_model_local=True,
            )


def test_oceanmodel_lon0_360():
    """Check for correct value of oceanmodel_lon0_360

    based on ocean model and lon input."""

    lon_in = -153

    m = ocean_model_simulation_mapper["CIOFSOP"](
        start_time="2022-01-01",
        lon=lon_in,
        lat=57,
        end_time="2022-01-01T05:00",
        ocean_model_local=True,
    )
    assert m.ocean_model_config.oceanmodel_lon0_360 == False
    assert m.lon == lon_in

    m = ocean_model_simulation_mapper["CIOFS"](
        start_time="2022-01-01",
        lon=lon_in,
        lat=57,
        end_time="2022-01-01T05:00",
        ocean_model_local=True,
    )
    assert m.ocean_model_config.oceanmodel_lon0_360 == False
    assert m.lon == lon_in

    m = ocean_model_simulation_mapper["CIOFSFRESH"](
        start_time="2004-01-01",
        lon=lon_in,
        lat=57,
        end_time="2004-01-01T05:00",
        ocean_model_local=True,
    )
    assert m.ocean_model_config.oceanmodel_lon0_360 == False
    assert m.lon == lon_in

    m = ocean_model_simulation_mapper["CIOFS3"](
        start_time="2004-01-01",
        lon=lon_in,
        lat=57,
        end_time="2004-01-01T05:00",
        ocean_model_local=True,
    )
    assert m.ocean_model_config.oceanmodel_lon0_360 == False
    assert m.lon == lon_in

    m = ocean_model_simulation_mapper["NWGOA"](
        start_time="2007-01-01",
        lon=lon_in,
        lat=57,
        end_time="2007-01-01T05:00",
        ocean_model_local=True,
    )
    assert m.ocean_model_config.oceanmodel_lon0_360 == True
    assert m.lon == lon_in + 360


def test_start_end_times():
    """Check for valid start_time and end_time values

    ...for specific ocean models of type OceanModelSimulation.
    Checked for general ranges in test_config_the_manager.py.
    """

    for ocean_model, test in tests_invalid.items():
        with pytest.raises(ValidationError):
            m = ocean_model_simulation_mapper[ocean_model](
                start_time=test["start_time"],
                end_time=tests_valid[ocean_model]["end_time"],
                lon=tests_valid[ocean_model]["lon"],
                lat=tests_valid[ocean_model]["lat"],
                ocean_model_local=True,
            )
        with pytest.raises(ValidationError):
            m = ocean_model_simulation_mapper[ocean_model](
                start_time=tests_valid[ocean_model]["start_time"],
                end_time=test["end_time"],
                lon=tests_valid[ocean_model]["lon"],
                lat=tests_valid[ocean_model]["lat"],
                ocean_model_local=True,
            )


class MockFileSystem:
    """Mock file system to simulate globbing for kerchunk JSON files."""

    def __init__(self, files):
        self.files = files

    def glob(self, year):
        # Mock glob method takes a year string instead of a pattern
        return [file for file in self.files if year in file]


MOCK_FILES_YEARLY_DATES = [
    (1, 1),
    (1, 2),
    (1, 31),
    (3, 15),
    (5, 31),
    (6, 1),
    (6, 15),
    (6, 30),
    (7, 1),
    (9, 15),
    (12, 1),
    (12, 30),
    (12, 31),
]
MOCK_FILES_DATES = [
    datetime(year, month, day)
    for month, day in MOCK_FILES_YEARLY_DATES
    for year in [2019, 2020, 2021]
]
TEST_DATE_RANGES = [
    ((2020, 6, 5), (2020, 6, 25), 1),
    ((2020, 1, 1), (2020, 1, 31), 3),
    ((2020, 1, 31), (2020, 1, 1), 3),
    ((2019, 12, 1), (2020, 1, 31), 6),
    ((2019, 12, 31), (2020, 1, 1), 2),
    ((2020, 1, 1), (2020, 12, 31), 13),
    ((2020, 1, 3), (2020, 6, 29), 5),
    ((2020, 6, 29), (2020, 1, 3), 5),
    ((2019, 12, 31), (2021, 1, 1), None),
    ((2019, 12, 31), (2021, 12, 31), None),
    ((2020, 1, 1), (2021, 12, 31), 26),
]


@pytest.mark.parametrize(
    "start_tuple, end_tuple, expected",
    TEST_DATE_RANGES,
    ids=[f"{start}->{end}" for start, end, _ in TEST_DATE_RANGES],
)
@pytest.mark.parametrize(
    "filename_format",
    ["%Y_0%j", "ciofs_%Y-%m-%d", "nwgoa_%Y-%m-%d"],
    ids=["CIOFS", "CIOFSOP", "NWGOA"],
)
def test_krchunk_json_filtering(filename_format, start_tuple, end_tuple, expected):
    """Check that kerchunk JSON files are correctly filtered based on date range"""
    start = datetime(*start_tuple)
    end = datetime(*end_tuple)

    all_files = [f"{d.strftime(filename_format)}.json" for d in MOCK_FILES_DATES]
    mock_fs = MockFileSystem(all_files)

    def make_glob_with_check(year: str) -> str:
        """Check that the input string is a year and return it instead of any pattern.

        The return value will be used by `MockFileSystem.glob` which accounts for this.
        """
        assert year == f"{start.year:04}" or year == f"{end.year:04}"
        return year

    try:
        jsons = find_json_files_in_date_range(
            mock_fs,
            make_glob_with_check,
            start,
            end,
            filename_format,
        )

        assert len(jsons) == expected
    except ValueError:
        assert expected is None, f"Expected {expected} JSON files but got ValueError"


def test_user_registry():

    # Create a temporary directory
    with tempfile.TemporaryDirectory() as temp_dir:
        # Define the content of the YAML file
        data = {
            "USER": {
                "name": "USER",
                "loc_remote": None,
                "temporal_resolution_str": "PT1H",
                "lon_min": -180,
                "lon_max": 180,
                "lat_min": -90,
                "lat_max": 90,
                "start_time_model": "2009-01-01T00:00",
                "end_time_fixed": "2010-01-01T00:00:00",
                "oceanmodel_lon0_360": False,
                "standard_name_mapping": {},
                "model_drop_vars": [],
                "dx": 1000,
                "kerchunk_func_str": None,
                "chunks": None,
            }
        }

        # Define the path for the YAML file
        yaml_file_path = Path(temp_dir) / f"USER.yaml"

        # Write the YAML data to the file
        with open(yaml_file_path, "w") as file:
            yaml.dump(data, file)

        # Set an environment variable for testing purposes
        # this defines where to find user templates with *.yaml
        os.environ["PTM_CONFIG_DIR"] = str(temp_dir)

        # Check if it's defined
        env_var = os.getenv("PTM_CONFIG_DIR")

        # reload particle_tracking_manager since we changed the environment variable
        importlib.reload(particle_tracking_manager.ocean_model_registry)

        # now USER is in the registry
        assert (
            "USER"
            in particle_tracking_manager.ocean_model_registry.ocean_model_registry.all()
        )


def test_onthefly_registry():

    ds_info = dict(
        temporal_resolution_str="PT1H",
        lon_min=1,
        lon_max=3,
        lat_min=1,
        lat_max=2,
        start_time_model=0,
        end_time_fixed=1,
    )
    particle_tracking_manager.config_ocean_model.register_on_the_fly(ds_info)

    # now new config is in the registry
    assert ocean_model_registry.get("ONTHEFLY").lon_min == 1


@patch("particle_tracking_manager.ocean_model_registry.calculate_CIOFSOP_max")
def test_CIOFSOP_max_update(model_CIOFSOP_max):
    """Make sure that CIOFSOP end_time_model is updated

    ...when the model is updated. Mock input from calculate_CIOFSOP_max.
    """

    # check the value once
    return_value1 = datetime(2025, 1, 16, 23, 0)
    model_CIOFSOP_max.return_value = return_value1
    m = particle_tracking_manager.OpenDriftModel(drift_model="Leeway", steps=1)
    assert m.config.ocean_model_config.end_time_model == return_value1

    # check the value again to see if it is updated, like it would be in real life
    return_value2 = datetime(2025, 2, 16, 23, 0)
    model_CIOFSOP_max.return_value = return_value2
    m = particle_tracking_manager.OpenDriftModel(drift_model="Leeway", steps=1)
    assert m.config.ocean_model_config.end_time_model == return_value2
